/*
 *
 *   a halafied local printer configuration app based on lpadmin from cups
 *
 *   Copyright 1997-2003 by Easy Software Products, all rights reserved.
 *   Copyright (C) 2004 Red Hat, Inc.
 *
 *   These coded instructions, statements, and computer programs are the
 *   property of Easy Software Products and are protected by Federal
 *   copyright law.  Distribution and use rights are outlined in the file
 *   "LICENSE" which should have been included with this file.  If this
 *   file is missing or damaged please contact Easy Software Products
 *   at:
 *
 *       Attn: CUPS Licensing Information
 *       Easy Software Products
 *       44141 Airport View Drive, Suite 204
 *       Hollywood, Maryland 20636-3111 USA
 *
 *       Voice: (301) 373-9603
 *       EMail: cups-info@cups.org
 *         WWW: http://www.cups.org
 *
 *   This file is subject to the Apple OS-Developed Software exception.
 *
 * Contents:
 *
 *   main()         - Send a file to the specified HAL device.
 *   list_devices() - List all HAL printer devices.
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <cups/cups.h>
#include <cups/language.h>
#include <config.h>
#ifdef HAVE_LIBZ
#  include <zlib.h>
#endif /* HAVE_LIBZ */
#include <hal/libhal.h>
#include <limits.h>
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>

#define URI_PREFIX "hal://"

/*
 * The debug macros are used if you compile with DEBUG defined.
 *
 * Usage:
 *
 *   DEBUG_puts("string")
 *   DEBUG_printf(("format string", arg, arg, ...));
 *
 * Note the extra parenthesis around the DEBUG_printf macro...
 */


/* Helper macros */

#  ifdef DEBUG
#    define DEBUG_puts(x) puts(x)
#    define DEBUG_printf(x) printf x
#  else
#    define DEBUG_puts(x)
#    define DEBUG_printf(x)
#  endif /* DEBUG */

/* change an occurance of character from to character to in string s */
void 
strchrep(char *s, char from, char to) 
{ 
      while ((s = strchr(s, from)) != NULL)
        *s = to;
}

static int	enable_printer(http_t *, char *);
static int	delete_printer(http_t *, char *);
static char	*get_line(char *, int, FILE *fp);
static int	set_printer_device(http_t *, char *, char *);
static int	set_printer_file(http_t *, char *, char *);
static int	set_printer_info(http_t *, char *, char *);
static int	set_printer_location(http_t *, char *, char *);
static int	set_printer_model(http_t *, char *, char *);
static int	set_printer_options(http_t *, char *, int, cups_option_t *);
static int	validate_name(char *);

#ifdef USE_PRINTCONF
static void	remove_printer_using_printconf(http_t *, const char *, const char *, const char *, const char *);
static int	set_printer_driver_using_printconf(http_t *, const char *, const char *, const char *, const char *, dbus_bool_t);
static int	printer_needs_config(http_t *, const char *, const char *, const char *, const char *, DBusError *);
#endif


static void	print_usage_and_exit();
static void	add_hal_device_if_printer(LibHalContext *, const char *);
static void	remove_hal_device_if_printer(LibHalContext *, const char *);

static char 	   *construct_printer_name(const char *, int);
static dbus_bool_t dbus_get_user_driver (DBusConnection *, const char *, const char *, const char *, const char *, char **, char **, DBusError *);
static void        dbus_prompt_printer_driver (DBusConnection *, const char *, const char *, const char *, const char *, DBusError *);

static void	sig_hup_cups();

/*
 * 'main()' - Parse options and configure the scheduler.
 */

int
main(int  argc,			/* I - Number of command-line arguments */
     char *argv[])		/* I - Command-line arguments */
{

  char		*hal_udi;	/* UDI of hal device*/  
  int		i;		/* Looping var */
  int		remove_device;  /* if true then we are removing the device, else adding */
  LibHalContext *hal_ctx;       /* Hal context */


  hal_udi     = NULL;
  remove_device = 0;

  for (i = 1; i < argc; i ++)
    if (argv[i][0] == '-')
      switch (argv[i][1])
      {
        case 'p' : /* device UID */
            if (argv[i][2])
            {
              hal_udi = argv[i] + 2;
            }
            else
            { 
               i ++;

	      if (i >= argc)
	      { 
                print_usage_and_exit();
	      }

	      hal_udi = argv[i];
            }
            break;

        case 'x' : /* remove device */
            remove_device = 1;

            if (argv[i][2])
            {
              hal_udi = argv[i] + 2;
            }
            else
            { 
               i ++;

	      if (i >= argc)
	      { 
                print_usage_and_exit();
	      }

	      hal_udi = argv[i];
            }
            break;
      }
  
  if (hal_udi == NULL)
    print_usage_and_exit();

  /* get hal context */
  if ((hal_ctx = hal_initialize(NULL, 0)) == NULL) 
  {
    fprintf(stderr, "hal_initialize failed\n");
    exit(1);
  } 

  if (!remove_device)  
    add_hal_device_if_printer(hal_ctx, hal_udi);
  else
    remove_hal_device_if_printer(hal_ctx, hal_udi);

  return (0);
}

/*
 * 'print_usage_and_exit()' - prints out the command line usage and exits
 */

static void
print_usage_and_exit()
{
  fprintf(stderr, "USAGE: hal_lpadmin -p <hal device udi> [-r]");
  exit(1);
}

/*
 * 'add_hal_device_if_printer()' - checks if the device is a printer and adds it to cups
 */
static void
add_hal_device_if_printer(LibHalContext *ctx, 
                          const char *udi)
{
  char *capabilities;
  char *product;
  char *vendor;
  char *description;
  char *serial;
  char *device;
  int usb_port_num;
  char *uri = NULL;
  char *printer = NULL;
  http_t *http = NULL;	

  capabilities = hal_device_get_property_string(ctx, udi, "info.capabilities");

  if (capabilities == NULL || strstr(capabilities, "printer") == NULL) 
  {
    hal_free_string(capabilities);
    return;
  }

  product = hal_device_get_property_string(ctx, udi, "printer.product");
  if (product == NULL)
  {
    product = hal_device_get_property_string(ctx, udi, "usb.product");
    if (product == NULL)
      product = strdup("Unknown");
  }

  vendor = hal_device_get_property_string(ctx, udi, "printer.vendor");
  if (vendor == NULL)
  {
    vendor = hal_device_get_property_string(ctx, udi, "usb.vendor");
    if (vendor == NULL)
      vendor = strdup("Unknown");
  }

  description = hal_device_get_property_string(ctx, udi, "printer.description");
  serial = hal_device_get_property_string(ctx, udi, "printer.serial");
  device = hal_device_get_property_string(ctx, udi, "printer.device");
  usb_port_num = hal_device_get_property_int(ctx, udi, "usb.port_number");

  uri = malloc(sizeof(char) * (strlen(udi) + strlen(URI_PREFIX) + 1));

  if (!uri) 
  {
    fprintf(stderr, "ERROR: Out of memory\n"); 
    goto out;
  }

  sprintf(uri, "%s%s", URI_PREFIX, udi); 
  
  printer = construct_printer_name(product, usb_port_num);

  if (!printer) 
    goto out;

  http = httpConnectEncrypt(cupsServer(), ippPort(),
                                   cupsEncryption());

  if (http == NULL)
  {
    fprintf(stderr, "Unable to connect to server\n");
    goto out;
  }

#ifndef USE_PRINTCONF
  enable_printer(http, printer);
  set_printer_device(http, printer, uri);
  set_printer_info(http, printer, description);

/* TODO: Break out printconf driver detection code */
#else 
  set_printer_driver_using_printconf(http, printer, vendor, product, udi, TRUE);
#endif

  out:
    hal_free_string(capabilities);
    hal_free_string(product);
    hal_free_string(vendor);
    hal_free_string(description);
    hal_free_string(serial);
    hal_free_string(device);
    hal_free_string(uri);
    hal_free_string(printer);
}

/*
 * 'remove_hal_device_if_printer()' - checks if the device is a printer and removes it from cups
 */
static void
remove_hal_device_if_printer(LibHalContext *ctx, 
                          const char *udi)
{
  char *capabilities;
  char *product;
  char *serial;
  char *vendor;
  int usb_port_num;
  char *printer = NULL;

  http_t *http = NULL;

  capabilities = hal_device_get_property_string(ctx, udi, "info.capabilities");

  if (capabilities == NULL || strstr(capabilities, "printer") == NULL)
  {
    hal_free_string(capabilities);
    return;
  }

  product = hal_device_get_property_string(ctx, udi, "printer.product");
  if (product == NULL)
  {
    product = hal_device_get_property_string(ctx, udi, "usb.product");
    if (product == NULL)
      product = strdup("Unknown");
  }

  vendor = hal_device_get_property_string(ctx, udi, "printer.vendor");
  if (vendor == NULL)
  {
    vendor = hal_device_get_property_string(ctx, udi, "usb.vendor");
    if (vendor == NULL)
      vendor = strdup("Unknown");
  }

  serial = hal_device_get_property_string(ctx, udi, "printer.serial");
  usb_port_num = hal_device_get_property_int(ctx, udi, "usb.port_number");

  printer = construct_printer_name(product, usb_port_num); 

  if (!printer) 
    goto out;

  http = httpConnectEncrypt(cupsServer(), ippPort(),
                                   cupsEncryption());

  if (http == NULL)
  {
    fprintf(stderr, "Unable to connect to server\n");
    goto out;
  }

#ifdef USE_PRINTCONF
  remove_printer_using_printconf(http, printer, vendor, product, udi);
#else
  delete_printer(http, printer);
#endif

  out:
    hal_free_string(capabilities);
    hal_free_string(product);
    hal_free_string(serial);
    hal_free_string(printer);

  
}


/*
 * 'delete_printer()' - Delete a printer from the system...
 */

static int				/* O - 0 on success, 1 on fail */
delete_printer(http_t *http,		/* I - Server connection */
               char   *printer)		/* I - Printer to delete */
{
  ipp_t		*request,		/* IPP Request */
		*response;		/* IPP Response */
  cups_lang_t	*language;		/* Default language */
  char		uri[HTTP_MAX_URI];	/* URI for printer/class */


  DEBUG_printf(("delete_printer(%p, \"%s\")\n", http, printer));

 /*
  * Build a CUPS_DELETE_PRINTER request, which requires the following
  * attributes:
  *
  *    attributes-charset
  *    attributes-natural-language
  *    printer-uri
  */

  snprintf(uri, sizeof(uri), "ipp://localhost/printers/%s", printer);

  request = ippNew();

  request->request.op.operation_id = CUPS_DELETE_PRINTER;
  request->request.op.request_id   = 1;

  language = cupsLangDefault();

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_CHARSET,
               "attributes-charset", NULL, cupsLangEncoding(language));

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_LANGUAGE,
               "attributes-natural-language", NULL, language->language);

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_URI,
               "printer-uri", NULL, uri);

 /*
  * Do the request and get back a response...
  */

  if ((response = cupsDoRequest(http, request, "/admin/")) == NULL)
  {
    fprintf(stderr, "lpadmin: delete-printer failed: %s\n",
            ippErrorString(cupsLastError()));

    return (1);
  }
  else if (response->request.status.status_code > IPP_OK_CONFLICT)
  {
    fprintf(stderr, "lpadmin: delete-printer failed: %s\n",
            ippErrorString(response->request.status.status_code));

    ippDelete(response);

    return (1);
  }
  else
  {
    ippDelete(response);

    return (0);
  }
}
/*
 * 'enable_printer()' - Enable a printer...
 */

static int				/* O - 0 on success, 1 on fail */
enable_printer(http_t *http,		/* I - Server connection */
               char   *printer)		/* I - Printer to enable */
{
  ipp_t		*request,		/* IPP Request */
		*response;		/* IPP Response */
  cups_lang_t	*language;		/* Default language */
  char		uri[HTTP_MAX_URI];	/* URI for printer/class */


  DEBUG_printf(("enable_printer(%p, \"%s\")\n", http, printer));

 /*
  * Build a CUPS_ADD_PRINTER request, which requires the following
  * attributes:
  *
  *    attributes-charset
  *    attributes-natural-language
  *    printer-uri
  *    printer-state
  *    printer-is-accepting-jobs
  */

  snprintf(uri, sizeof(uri), "ipp://localhost/printers/%s", printer);

  request = ippNew();

  request->request.op.operation_id = CUPS_ADD_PRINTER;
  request->request.op.request_id   = 1;

  language = cupsLangDefault();

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_CHARSET,
               "attributes-charset", NULL, cupsLangEncoding(language));

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_LANGUAGE,
               "attributes-natural-language", NULL, language->language);

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_URI,
               "printer-uri", NULL, uri);

  ippAddInteger(request, IPP_TAG_PRINTER, IPP_TAG_ENUM, "printer-state",
                IPP_PRINTER_IDLE);

  ippAddBoolean(request, IPP_TAG_PRINTER, "printer-is-accepting-jobs", 1);

 /*
  * Do the request and get back a response...
  */

  if ((response = cupsDoRequest(http, request, "/admin/")) == NULL)
  {
    fprintf(stderr, "lpadmin: add-printer (enable) failed: %s\n",
            ippErrorString(cupsLastError()));
    return (1);
  }
  else if (response->request.status.status_code > IPP_OK_CONFLICT)
  {
    fprintf(stderr, "lpadmin: add-printer (enable) failed: %s\n",
            ippErrorString(response->request.status.status_code));

    ippDelete(response);

    return (1);
  }
  else
  {
    ippDelete(response);

    return (0);
  }
}


/*
 * 'get_line()' - Get a line that is terminated by a LF, CR, or CR LF.
 */

static char *				/* O - Pointer to buf or NULL on EOF */
get_line(char *buf,			/* I - Line buffer */
         int  length,			/* I - Length of buffer */
	 FILE *fp)			/* I - File to read from */
{
  char	*bufptr;			/* Pointer into buffer */
  int	ch;				/* Character from file */


  length --;
  bufptr = buf;

  while ((ch = getc(fp)) != EOF)
  {
    if (ch == '\n')
      break;
    else if (ch == '\r')
    {
     /*
      * Look for LF...
      */

      ch = getc(fp);
      if (ch != '\n' && ch != EOF)
        ungetc(ch, fp);

      break;
    }

    *bufptr++ = ch;
    length --;
    if (length == 0)
      break;
  }

  *bufptr = '\0';

  if (ch == EOF)
    return (NULL);
  else
    return (buf);
}


/*
 * 'set_printer_device()' - Set the device-uri attribute.
 */

static int				/* O - 0 on success, 1 on fail */
set_printer_device(http_t *http,	/* I - Server connection */
                   char   *printer,	/* I - Printer */
		   char   *device)	/* I - New device URI */
{
  ipp_t		*request,		/* IPP Request */
		*response;		/* IPP Response */
  cups_lang_t	*language;		/* Default language */
  char		uri[HTTP_MAX_URI];	/* URI for printer/class */


  DEBUG_printf(("set_printer_device(%p, \"%s\", \"%s\")\n", http, printer,
                device));

 /*
  * Build a CUPS_ADD_PRINTER request, which requires the following
  * attributes:
  *
  *    attributes-charset
  *    attributes-natural-language
  *    printer-uri
  */

  snprintf(uri, sizeof(uri), "ipp://localhost/printers/%s", printer);

  request = ippNew();

  request->request.op.operation_id = CUPS_ADD_PRINTER;
  request->request.op.request_id   = 1;

  language = cupsLangDefault();

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_CHARSET,
               "attributes-charset", NULL, cupsLangEncoding(language));

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_LANGUAGE,
               "attributes-natural-language", NULL, language->language);

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_URI,
               "printer-uri", NULL, uri);

 /*
  * Add the device URI...
  */

  if (device[0] == '/')
  {
   /*
    * Convert filename to URI...
    */

    snprintf(uri, sizeof(uri), "file:%s", device);
    ippAddString(request, IPP_TAG_PRINTER, IPP_TAG_URI, "device-uri", NULL,
                 uri);

    printf("%s\n",uri);
  }
  else
  {
    ippAddString(request, IPP_TAG_PRINTER, IPP_TAG_URI, "device-uri", NULL,
                 device);
    printf("%s\n",device);
  }
 /*
  * Do the request and get back a response...
  */

  if ((response = cupsDoRequest(http, request, "/admin/")) == NULL)
  {
    fprintf(stderr, "lpadmin: add-printer (set device) failed: %s\n",
            ippErrorString(cupsLastError()));
    return (1);
  }
  else if (response->request.status.status_code > IPP_OK_CONFLICT)
  {
    fprintf(stderr, "lpadmin: add-printer (set device) failed: %s\n",
            ippErrorString(response->request.status.status_code));

    ippDelete(response);

    return (1);
  }
  else
  {
    ippDelete(response);

    return (0);
  }
}


/*
 * 'set_printer_file()' - Set the interface script or PPD file.
 */

static int				/* O - 0 on success, 1 on fail */
set_printer_file(http_t *http,		/* I - Server connection */
                 char   *printer,	/* I - Printer */
		 char   *file)		/* I - PPD file or interface script */
{
  ipp_status_t	status;			/* IPP status code */
  ipp_t		*request,		/* IPP Request */
		*response;		/* IPP Response */
  cups_lang_t	*language;		/* Default language */
  char		uri[HTTP_MAX_URI];	/* URI for printer/class */
#ifdef HAVE_LIBZ
  char		tempfile[1024];		/* Temporary filename */
  int		fd;			/* Temporary file */
  gzFile	*gz;			/* GZIP'd file */
  char		buffer[8192];		/* Copy buffer */
  int		bytes;			/* Bytes in buffer */


  DEBUG_printf(("set_printer_file(%p, \"%s\", \"%s\")\n", http, printer,
                file));

 /*
  * See if the file is gzip'd; if so, unzip it to a temporary file and
  * send the uncompressed file.
  */

  if (strcmp(file + strlen(file) - 3, ".gz") == 0)
  {
   /*
    * Yes, the file is compressed; uncompress to a temp file...
    */

    if ((fd = cupsTempFd(tempfile, sizeof(tempfile))) < 0)
    {
      perror("lpadmin: Unable to create temporary file");
      return (1);
    }

    if ((gz = gzopen(file, "rb")) == NULL)
    {
      perror("lpadmin: Unable to open file");
      close(fd);
      unlink(tempfile);
      return (1);
    }

    while ((bytes = gzread(gz, buffer, sizeof(buffer))) > 0)
      write(fd, buffer, bytes);

    close(fd);
    gzclose(gz);

    file = tempfile;
  }
#endif /* HAVE_LIBZ */

 /*
  * Build a CUPS_ADD_PRINTER request, which requires the following
  * attributes:
  *
  *    attributes-charset
  *    attributes-natural-language
  *    printer-uri
  */

  snprintf(uri, sizeof(uri), "ipp://localhost/printers/%s", printer);

  request = ippNew();

  request->request.op.operation_id = CUPS_ADD_PRINTER;
  request->request.op.request_id   = 1;

  language = cupsLangDefault();

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_CHARSET,
               "attributes-charset", NULL, cupsLangEncoding(language));

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_LANGUAGE,
               "attributes-natural-language", NULL, language->language);

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_URI,
               "printer-uri", NULL, uri);

 /*
  * Do the request and get back a response...
  */

  if ((response = cupsDoFileRequest(http, request, "/admin/", file)) == NULL)
    status = cupsLastError();
  else
  {
    status = response->request.status.status_code;
    ippDelete(response);
  }

#ifdef HAVE_LIBZ
 /*
  * Remove the temporary file as needed...
  */

  if (file == tempfile)
    unlink(tempfile);
#endif /* HAVE_LIBZ */

  if (status > IPP_OK_CONFLICT)
  {
    fprintf(stderr, "lpadmin: add-printer (set model) failed: %s\n",
            ippErrorString(status));

    return (1);
  }
  else
    return (0);
}


/*
 * 'set_printer_info()' - Set the printer description string.
 */

static int				/* O - 0 on success, 1 on fail */
set_printer_info(http_t *http,		/* I - Server connection */
                 char   *printer,	/* I - Printer */
		 char   *info)		/* I - New description string */
{
  ipp_t		*request,		/* IPP Request */
		*response;		/* IPP Response */
  cups_lang_t	*language;		/* Default language */
  char		uri[HTTP_MAX_URI];	/* URI for printer/class */


  DEBUG_printf(("set_printer_info(%p, \"%s\", \"%s\")\n", http, printer,
                info));

 /*
  * Build a CUPS_ADD_PRINTER request, which requires the following
  * attributes:
  *
  *    attributes-charset
  *    attributes-natural-language
  *    printer-uri
  */

  snprintf(uri, sizeof(uri), "ipp://localhost/printers/%s", printer);

  request = ippNew();

  request->request.op.operation_id = CUPS_ADD_PRINTER;
  request->request.op.request_id   = 1;

  language = cupsLangDefault();

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_CHARSET,
               "attributes-charset", NULL, cupsLangEncoding(language));

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_LANGUAGE,
               "attributes-natural-language", NULL, language->language);

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_URI,
               "printer-uri", NULL, uri);

 /*
  * Add the info string...
  */

  ippAddString(request, IPP_TAG_PRINTER, IPP_TAG_TEXT, "printer-info", NULL,
               info);

 /*
  * Do the request and get back a response...
  */

  if ((response = cupsDoRequest(http, request, "/admin/")) == NULL)
  {
    fprintf(stderr, "lpadmin: add-printer (set description) failed: %s\n",
            ippErrorString(cupsLastError()));
    return (1);
  }
  else if (response->request.status.status_code > IPP_OK_CONFLICT)
  {
    fprintf(stderr, "lpadmin: add-printer (set description) failed: %s\n",
            ippErrorString(response->request.status.status_code));

    ippDelete(response);

    return (1);
  }
  else
  {
    ippDelete(response);

    return (0);
  }
}


/*
 * 'set_printer_location()' - Set the printer location string.
 */

static int				/* O - 0 on success, 1 on fail */
set_printer_location(http_t *http,	/* I - Server connection */
                     char   *printer,	/* I - Printer */
		     char   *location)	/* I - New location string */
{
  ipp_t		*request,		/* IPP Request */
		*response;		/* IPP Response */
  cups_lang_t	*language;		/* Default language */
  char		uri[HTTP_MAX_URI];	/* URI for printer/class */


  DEBUG_printf(("set_printer_location(%p, \"%s\", \"%s\")\n", http, printer,
                location));

 /*
  * Build a CUPS_ADD_PRINTER request, which requires the following
  * attributes:
  *
  *    attributes-charset
  *    attributes-natural-language
  *    printer-uri
  */

  snprintf(uri, sizeof(uri), "ipp://localhost/printers/%s", printer);

  request = ippNew();

  request->request.op.operation_id = CUPS_ADD_PRINTER;
  request->request.op.request_id   = 1;

  language = cupsLangDefault();

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_CHARSET,
               "attributes-charset", NULL, cupsLangEncoding(language));

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_LANGUAGE,
               "attributes-natural-language", NULL, language->language);

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_URI,
               "printer-uri", NULL, uri);

 /*
  * Add the location string...
  */

  ippAddString(request, IPP_TAG_PRINTER, IPP_TAG_TEXT, "printer-location", NULL,
               location);

 /*
  * Do the request and get back a response...
  */

  if ((response = cupsDoRequest(http, request, "/admin/")) == NULL)
  {
    fprintf(stderr, "lpadmin: add-printer (set location) failed: %s\n",
            ippErrorString(cupsLastError()));

    return (1);
  }
  else if (response->request.status.status_code > IPP_OK_CONFLICT)
  {
    fprintf(stderr, "lpadmin: add-printer (set location) failed: %s\n",
            ippErrorString(response->request.status.status_code));

    ippDelete(response);

    return (1);
  }
  else
  {
    ippDelete(response);

    return (0);
  }
}


/*
 * 'set_printer_model()' - Set the driver model file.
 */

static int				/* O - 0 on success, 1 on fail */
set_printer_model(http_t *http,		/* I - Server connection */
                  char   *printer,	/* I - Printer */
		  char   *model)	/* I - Driver model file */
{
  ipp_t		*request,		/* IPP Request */
		*response;		/* IPP Response */
  cups_lang_t	*language;		/* Default language */
  char		uri[HTTP_MAX_URI];	/* URI for printer/class */


 /*
  * Build a CUPS_ADD_PRINTER request, which requires the following
  * attributes:
  *
  *    attributes-charset
  *    attributes-natural-language
  *    printer-uri
  *    ppd-name
  */

  snprintf(uri, sizeof(uri), "ipp://localhost/printers/%s", printer);

  request = ippNew();

  request->request.op.operation_id = CUPS_ADD_PRINTER;
  request->request.op.request_id   = 1;

  language = cupsLangDefault();

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_CHARSET,
               "attributes-charset", NULL, cupsLangEncoding(language));

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_LANGUAGE,
               "attributes-natural-language", NULL, language->language);

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_URI,
               "printer-uri", NULL, uri);

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_NAME,
               "ppd-name", NULL, model);

 /*
  * Do the request and get back a response...
  */

  if ((response = cupsDoRequest(http, request, "/admin/")) == NULL)
  {
    fprintf(stderr, "lpadmin: add-printer (set model) failed: %s\n",
            ippErrorString(cupsLastError()));

    return (1);
  }
  else if (response->request.status.status_code > IPP_OK_CONFLICT)
  {
    fprintf(stderr, "lpadmin: add-printer (set model) failed: %s\n",
            ippErrorString(response->request.status.status_code));

    ippDelete(response);

    return (1);
  }
  else
  {
    ippDelete(response);

    return (0);
  }
}


/*
 * 'set_printer_options()' - Set the printer options.
 */

static int				/* O - 0 on success, 1 on fail */
set_printer_options(http_t        *http,/* I - Server connection */
                    char          *printer,
		    			/* I - Printer */
		    int           num_options,
		    			/* I - Number of options */
		    cups_option_t *options)
		    			/* I - Options */
{
  ipp_t		*request,		/* IPP Request */
		*response;		/* IPP Response */
  ipp_attribute_t *attr;		/* IPP attribute */
  cups_lang_t	*language;		/* Default language */
  ipp_op_t	op;			/* Operation to perform */
  const char	*val,			/* Option value */
		*ppdfile;		/* PPD filename */
  char		uri[HTTP_MAX_URI],	/* URI for printer/class */
		line[1024],		/* Line from PPD file */
		keyword[1024],		/* Keyword from Default line */
		*keyptr,		/* Pointer into keyword... */
		tempfile[1024];		/* Temporary filename */
  FILE		*in,			/* PPD file */
		*out;			/* Temporary file */
  int		outfd;			/* Temporary file descriptor */
  const char	*protocol;		/* Protocol */


  DEBUG_printf(("set_printer_options(%p, \"%s\", %d, %p)\n", http, printer,
                num_options, options));

  language = cupsLangDefault();

  snprintf(uri, sizeof(uri), "ipp://localhost/printers/%s", printer);

 /*
  * Build a GET_PRINTER_ATTRIBUTES request, which requires the following
  * attributes:
  *
  *    attributes-charset
  *    attributes-natural-language
  *    printer-uri
  *    requested-attributes
  */

  request = ippNew();

  request->request.op.operation_id = IPP_GET_PRINTER_ATTRIBUTES;
  request->request.op.request_id   = 1;

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_CHARSET,
               "attributes-charset", NULL, cupsLangEncoding(language));

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_LANGUAGE,
               "attributes-natural-language", NULL, language->language);

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_URI,
               "printer-uri", NULL, uri);

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_KEYWORD,
               "requested-attributes", NULL, "printer-type");

 /*
  * Do the request...
  */

  op = CUPS_ADD_PRINTER;

  if ((response = cupsDoRequest(http, request, "/")) != NULL)
  {
   /*
    * See what kind of printer or class it is...
    */

    if ((attr = ippFindAttribute(response, "printer-type", IPP_TAG_ENUM)) != NULL)
    {
      if (attr->values[0].integer & (CUPS_PRINTER_CLASS | CUPS_PRINTER_IMPLICIT))
      {
        op = CUPS_ADD_CLASS;
        snprintf(uri, sizeof(uri), "ipp://localhost/classes/%s", printer);
      }
    }

    ippDelete(response);
  }

 /*
  * Build a CUPS_ADD_PRINTER or CUPS_ADD_CLASS request, which requires
  * the following attributes:
  *
  *    attributes-charset
  *    attributes-natural-language
  *    printer-uri
  *    other options
  */

  request = ippNew();

  request->request.op.operation_id = op;
  request->request.op.request_id   = 1;

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_CHARSET,
               "attributes-charset", NULL, cupsLangEncoding(language));

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_LANGUAGE,
               "attributes-natural-language", NULL, language->language);

  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_URI,
               "printer-uri", NULL, uri);

 /*
  * Add the options...
  */

  cupsEncodeOptions(request, num_options, options);

  if (op == CUPS_ADD_PRINTER)
    ppdfile = cupsGetPPD(printer);
  else
    ppdfile = NULL;

  if (ppdfile != NULL)
  {
   /*
    * Set default options in the PPD file...
    */

    if ((outfd = cupsTempFd(tempfile, sizeof(tempfile))) < 0)
    {
      fprintf(stderr, "lpadmin: Unable to create temporary file - %s\n",
              strerror(errno));
      ippDelete(request);
      unlink(ppdfile);
      return (1);
    }

    if ((in = fopen(ppdfile, "rb")) == NULL)
    {
      fprintf(stderr, "lpadmin: Unable to open PPD file \"%s\" - %s\n",
              ppdfile, strerror(errno));
      ippDelete(request);
      unlink(ppdfile);
      close(outfd);
      unlink(tempfile);
      return (1);
    }

    out      = fdopen(outfd, "wb");
    protocol = cupsGetOption("protocol", num_options, options);    

    while (get_line(line, sizeof(line), in) != NULL)
    {
      if (!strncmp(line, "*cupsProtocol:", 14) && protocol)
      {
       /*
        * Set a new output protocol (BCP or TBCP) below...
	*/

        continue;
      }
      else if (strncmp(line, "*Default", 8))
        fprintf(out, "%s\n", line);
      else
      {
       /*
        * Get default option name...
	*/

        strncpy(keyword, line + 8, sizeof(keyword));

	for (keyptr = keyword; *keyptr; keyptr ++)
	  if (*keyptr == ':' || isspace(*keyptr))
	    break;

        *keyptr = '\0';

        if (strcmp(keyword, "PageRegion") == 0)
	  val = cupsGetOption("PageSize", num_options, options);
	else
	  val = cupsGetOption(keyword, num_options, options);

        if (val != NULL)
	  fprintf(out, "*Default%s: %s\n", keyword, val);
	else
	  fprintf(out, "%s\n", line);
      }
    }

    if (protocol)
      fprintf(out, "*cupsProtocol: \"%s\"\n", protocol);

    fclose(in);
    fclose(out);
    close(outfd);

   /*
    * Do the request...
    */

    response = cupsDoFileRequest(http, request, "/admin/", tempfile);

   /*
    * Clean up temp files... (TODO: catch signals in case we CTRL-C during
    * lpadmin)
    */

    unlink(ppdfile);
    unlink(tempfile);
  }
  else
  {
   /*
    * No PPD file - just set the options...
    */

    response = cupsDoRequest(http, request, "/admin/");
  }

 /*
  * Check the response...
  */

  if (response == NULL)
  {
    fprintf(stderr, "lpadmin: %s failed: %s\n",
            op == CUPS_ADD_PRINTER ? "add-printer" : "add-class",
            ippErrorString(cupsLastError()));

    return (1);
  }
  else if (response->request.status.status_code > IPP_OK_CONFLICT)
  {
    fprintf(stderr, "lpadmin: %s failed: %s\n",
            op == CUPS_ADD_PRINTER ? "add-printer" : "add-class",
            ippErrorString(response->request.status.status_code));

    ippDelete(response);

    return (1);
  }
  else
  {
    ippDelete(response);

    return (0);
  }
}


/*
 * 'validate_name()' - Make sure the printer name only contains valid chars,
 *                     replace spaces with dashes and make all letters lowercase.
 */

static int			/* O - 0 if name is no good, 1 if name is good */
validate_name(char *name)	/* I - Name to check */
{
  char	*ptr;		/* Pointer into name */


 /*
  * Scan the whole name...
  */

  for (ptr = name; *ptr; ptr ++)
    if (*ptr == '@')
      break;
    else if (*ptr == ' ')
      *ptr = '-';
    else if ((*ptr >= 0 && *ptr < ' ') || *ptr == 127 || *ptr == '/')
      return (0);
    else
      *ptr = tolower(*ptr);

 /*
  * All the characters are good; validate the length, too...
  */

  return ((ptr - name) < 128);
}

/*
 * 'construct_printer_name()' -  creates a name by taking the product name
 *                               and appending it with the usb port number 
 *                    
 */
static char *
construct_printer_name(const char *product, int usb_port_num)
{
  char *printer;
  char port_num[256];

  snprintf(port_num, 256, "-%i", usb_port_num);
  printer = malloc(sizeof(char) * (strlen(port_num) + strlen(product) + 1)); 

  if (!printer) 
  {
    fprintf(stderr, "ERROR: Out of memory\n"); 
    return NULL;
  }

  sprintf(printer, "%s%s", product, port_num);

  if (!validate_name(printer)) 
  {
    fprintf(stderr, "Printer name %s, contains non printable characters\n", printer); 
    free(printer);
    return NULL;
  }
 
  return printer; 
}

#ifdef USE_PRINTCONF

static char *
translate_make(char *make)
{
  char *uc_make;
  int make_len, i;
  char *result;

  make_len = strlen(make);
  uc_make = strdup(make);

  for(i=0; i<make_len; i++)
  {
    uc_make[i] = toupper(uc_make[i]);
  }

  if (strcmp(uc_make, "HEWLETT-PACKARD") == 0)
    result = strdup("HP");
  else
    result = strdup(make);

  free (uc_make);
  return result;
}

char *
readlink_malloc (const char *filename)
{
  int size = 100;
  char *buffer = NULL;

  while (1)
  {
    buffer = (char *) realloc (buffer, size);
    int nchars = readlink (filename, buffer, size);
    if (nchars < 0)
    {
      free (buffer);
      return NULL;
    }
    if (nchars < size)
      return buffer;
    size *= 2;
  }
}

static void
sig_hup_cups ()
{
  FILE *cups_pid_file;
  char cups_pid_buf[256];
  int cups_pid;

  cups_pid_file = fopen ("/var/run/cupsd.pid", "r");

  if (cups_pid_file != NULL) 
  {
    fgets (cups_pid_buf, sizeof (cups_pid_buf), cups_pid_file);
    cups_pid = atoi (cups_pid_buf);
    fclose (cups_pid_file);
  } else
    goto fail;

  if (cups_pid > 0)
  {
    char proc_pid_file_name[265];
    char *pids_exe_name;

    snprintf (proc_pid_file_name, 265, "/proc/%s/exe", cups_pid_buf);

    pids_exe_name = readlink_malloc (proc_pid_file_name);

    if (pids_exe_name != NULL) 
    {
      if (strcmp (pids_exe_name, "/usr/sbin/cupsd") == 0)
      {
        kill (cups_pid, SIGHUP);
        return;
      }
      free (pids_exe_name);
    }
  }

  fail:
    system ("/usr/bin/killall -SIGHUP cupsd");
}

/*TODO: This is all just hacked together
        eventualy we want to break out the
        printconf stuff and make it saner*/

void
remove_printer_using_printconf(http_t *http, 
                               const char *printer, 
                               const char *vendor, 
                               const char *model,
                               const char *udi)
{
  int vendor_len, model_len, cmd_len, device_len;
  int remove_len;
  char vendor_switch[] = " --make=";
  char model_switch[] = " --model=";
  char device_switch[] = " --device=";
  char device_proto[] = "hal://"; 
  char remove_cmd[] = " --Xremove-local";
  char *printconf;

  vendor_len = strlen(vendor) + strlen(vendor_switch) + 2;
  model_len = strlen(model) + strlen(model_switch) + 2;
  device_len = strlen(udi) + strlen(device_switch) + strlen(device_proto) + 2;

  /*match command length*/
  cmd_len = strlen(remove_cmd);
  remove_len = vendor_len + model_len + cmd_len + 1;

  /* make sure no one tries to unquote untrusted strings */
  strchrep(vendor, '\"', '_');
  strchrep(model, '\"', '_');
  strchrep(model, '\\', '_');
  strchrep(vendor, '\\', '_');

  printconf = malloc(sizeof(char) * (remove_len + strlen(PRINTCONF_EXEC)));

  if (printconf == NULL) 
  {
    fprintf(stderr, "Out of memory error. Exiting!");
    exit(-1);
  }

  sprintf(printconf, "%s%s%s\"%s\"%s\"%s\"%s\"%s%s\"", PRINTCONF_EXEC, 
                                                       remove_cmd, 
                                                       vendor_switch, vendor, 
                                                       model_switch, model,
						       device_switch, device_proto, udi);

  system(printconf);
  system("/usr/sbin/printconf-backend --force-rebuild");

  sig_hup_cups ();

  free(printconf); 
}



static int
set_printer_driver_using_printconf(http_t *http, 
                                   const char *printer, 
                                   const char *vendor, 
                                   const char *model,
                                   const char *udi,
                                   dbus_bool_t prompt_session_for_driver)
{
  FILE *driver_pipe;
  char driver[PATH_MAX + 1];
  size_t chars_read;
  int vendor_len, model_len, cmd_len, name_len, device_len;
  int match_len, add_len;
  char vendor_switch[] = " --make=";
  char model_switch[] = " --model=";
  char name_switch[] = " --name=";
  char device_switch[] = " --device=";
  char device_proto[] = "hal://"; 
  char match_cmd[] = " --Xmatch-driver";
  char add_cmd[] = " --Xadd-local";
  char *printconf;
  char *make;
  int result;

  make = translate_make(vendor);

  vendor_len = strlen(make) + strlen(vendor_switch) + 2;
  model_len = strlen(model) + strlen(model_switch) + 2;
  name_len = strlen(printer) + strlen(name_switch) + 2;
  device_len = strlen(udi) + strlen(device_switch) + strlen(device_proto) + 2;

  /*match command length*/
  cmd_len = strlen(match_cmd);
  match_len = vendor_len + model_len + cmd_len + 1;

  /*add command length*/
  cmd_len = strlen(add_cmd);
  add_len = vendor_len + name_len + model_len + cmd_len + device_len + 1;


  /* make sure no one tries to unquote untrusted strings */
  strchrep(make, '\"', '_');
  strchrep(model, '\"', '_');
  strchrep(model, '\\', '_');
  strchrep(make, '\\', '_');

  printconf = malloc(sizeof(char) * (match_len + strlen(PRINTCONF_EXEC)));

  if (printconf == NULL) 
  {
    fprintf(stderr, "Out of memory error. Exiting!");
    exit(-1);
  }

  sprintf(printconf, "%s%s%s\"%s\"%s\"%s\"", PRINTCONF_EXEC, 
                                             match_cmd, 
                                             vendor_switch, make, 
                                             model_switch, model);

  driver_pipe = popen(printconf, "r");

  chars_read = fread(driver, sizeof(char), PATH_MAX, driver_pipe);
  driver[chars_read] = '\0';

  result = pclose(driver_pipe);

  free(printconf);

  if (result == 0)
  {
    printconf = malloc(sizeof(char) * (add_len + strlen(PRINTCONF_EXEC)));

    if (printconf == NULL) 
    {
      fprintf(stderr, "Out of memory error. Exiting!");
      exit(-1);
    }

    sprintf(printconf, "%s%s%s\"%s\"%s\"%s\"%s\"%s\"%s\"%s%s\"", PRINTCONF_EXEC, 
                                                       add_cmd, 
                                                       vendor_switch, make, 
                                                       model_switch, model,
                                                       name_switch, printer,
						       device_switch, device_proto, udi);

    system(printconf);
    system("/usr/sbin/printconf-backend --force-rebuild");

    sig_hup_cups ();
    free(printconf); 
  }
  else if (prompt_session_for_driver)
  {
    DBusError error;

    dbus_error_init(&error);
    if (printer_needs_config(http, vendor, model, udi, printer, &error) != 0) 
    {
      if (dbus_error_is_set (&error))
      {
        fprintf (stderr, "An error occurred while calling printer_needs_config: %s\n", error.message);
        dbus_error_free (&error);
      }	
    }
  }

  free(make);
  return result;

}

static void
dbus_prompt_printer_driver (DBusConnection *connection,
                            const char *make,
                            const char *model,
                            const char *udi,
                            const char *printer,
                            DBusError *error)
{
  DBusMessage *message;
  DBusMessageIter iter;

  message = dbus_message_new_method_call ("com.redhat.PrintDriverSelection",
                                          "/com/redhat/PrintDriverSelection",
                                          "com.redhat.PrintDriverSelection",
                                          "PromptPrintDriver");
                                                                                                                             
  dbus_message_append_iter_init (message, &iter);
  dbus_message_iter_append_string (&iter, make);         /*make*/
  dbus_message_iter_append_string (&iter, model);        /*model*/
  dbus_message_iter_append_string (&iter, udi);          /*printer udi*/
  dbus_message_iter_append_string (&iter, printer);      /*printer name*/
                                                                                                                             
  dbus_connection_send (connection, message, NULL);
  dbus_connection_flush (connection);
  dbus_message_unref (message);
}

dbus_bool_t 
dbus_get_user_driver (DBusConnection *connection,
                      const char *make,
                      const char *model,
                      const char *udi,
                      const char *printer,
                      char **return_make,
                      char **return_model,
                      DBusError *error)
{
  DBusMessage *message;
  DBusMessageIter iter;
  DBusMessage *reply_message;

  message = dbus_message_new_method_call ("com.redhat.PrintDriverSelection",
                                          "/com/redhat/PrintDriverSelection",
                                          "com.redhat.PrintDriverSelection",
                                          "GetUserDriver");
                                                                                                                             
  dbus_message_append_iter_init (message, &iter);
  dbus_message_iter_append_string (&iter, make);         /*make*/
  dbus_message_iter_append_string (&iter, model);        /*model*/
  dbus_message_iter_append_string (&iter, udi);          /*printer udi*/
  dbus_message_iter_append_string (&iter, printer);      /*printer name*/
                                                                                                                             
  reply_message = dbus_connection_send_with_reply_and_block (connection, message, 10000, error);

  if (dbus_error_is_set (error))
    return FALSE;

  if (!dbus_message_get_args (reply_message, error,
                              DBUS_TYPE_STRING, return_make, 
                              DBUS_TYPE_STRING, return_model,
                              DBUS_TYPE_INVALID))
  {
    *return_make = NULL;
    *return_model = NULL;
    dbus_error_free (error);
    return FALSE;
  }

  dbus_connection_flush (connection);
  dbus_message_unref (reply_message);
  dbus_message_unref (message);

  return TRUE;
}


static int
printer_needs_config(http_t *http,
                     const char *make, 
                     const char *model, 
                     const char *udi, 
                     const char *printer, 
                     DBusError *error)
{
  DBusConnection *connection;
  dbus_bool_t result;
  char *return_make;
  char *return_model;

  connection = dbus_bus_get (DBUS_BUS_SYSTEM, error);
  if (connection == NULL) 
  {
    return -1;
  }

  result = dbus_get_user_driver (connection, make, model, udi, printer, &return_make, &return_model, error);

  if (!result)
  {
    if (dbus_error_is_set (error))
      return -1;

    dbus_prompt_printer_driver (connection, make, model, udi, printer, error);
  }
  else
  {
    set_printer_driver_using_printconf (http, printer, return_make, return_model, udi, FALSE);

    if (return_make != NULL)
      free (return_make);
    if (return_model != NULL)
      free (return_model);
  }

  if (dbus_error_is_set (error))
      return -1;

  dbus_connection_unref (connection);

  return 0;
}

#endif
